/*
** File: adb3_platform.c  
** Project: ADB3 API module library
** Purpose: Implements Linux-specific functions for ADB3 API library
**
** (C) Copyright Alpha Data 2011
*/

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>

#include "adb3_platform.h"
#include <ioctl_dflinux.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const char g_deviceName[] = "adb3i";

#define ERRORS_ON_STDERR 0

ADB3_STATUS
adb3OpenDevice(
  unsigned int index,
  bool_t bPassive,
  ADB3_HANDLE* phAdb3)
{
  char filename[32];
  int hDevice;

  snprintf(filename, ARRAY_LENGTH(filename), "/dev/%s%u", g_deviceName, index);
  hDevice = open(filename, bPassive ? O_RDONLY : O_RDWR);
  if (hDevice < 0) {
    switch (errno) {
    case EACCES:
      return ADB3_ACCESS_DENIED;

    case ENXIO:
    case ENOENT:
      return ADB3_DEVICE_NOT_FOUND;

    default:
#if ERRORS_ON_STDERR
      fprintf(stderr, "adb3OpenDevice: errno=%d (%s)\n", errno, strerror(errno));
#endif
      return ADB3_UNEXPECTED_ERROR;
    }
  }
  *phAdb3 = hDevice;
  return ADB3_SUCCESS;
}

ADB3_STATUS
adb3CloseDevice(
  ADB3_HANDLE hAdb3)
{
  ADB3_STATUS status = ADB3_SUCCESS;

  /* Close the device handle. This prevents client code from using the handle to do further operations (such as mmap) */
  if (close(hAdb3) < 0) {
    switch (errno) {
    case EBADF:
      status = ADB3_INVALID_HANDLE;
      break;

    default:
#if ERRORS_ON_STDERR
      fprintf(stderr, "adb3CloseDevice: errno=%d (%s)\n", errno, strerror(errno));
#endif
      status = ADB3_UNEXPECTED_ERROR;
      break;
    }
  }

  return status;
}

#define MAX_IOCTL_SIZE (64)
#define IOCTL_SIZE_IN_SHIFT  (0)
#define IOCTL_SIZE_OUT_SHIFT (6)

ADB3_STATUS
adb3Ioctl(
  ADB3_HANDLE hAdb3,
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  if (sizeIn >= MAX_IOCTL_SIZE || sizeOut >= MAX_IOCTL_SIZE) {
#if ERRORS_ON_STDERR
    fprintf(stderr, "adb3Ioctl: sizeIn (%u) or sizeOut (%u) greater than 64\n", sizeIn, sizeOut);
#endif
    return ADB3_UNEXPECTED_ERROR;
  }
  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)code, ((int)sizeIn << IOCTL_SIZE_IN_SHIFT) | ((int)sizeOut << IOCTL_SIZE_OUT_SHIFT));
  err = ioctl(hAdb3, ioctlCode, pIoctl);
  if (0 == err) {
    return ADB3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADB3_INVALID_HANDLE;

      case ENOMEM:
        return ADB3_NO_MEMORY;

      case ECANCELED:
        return ADB3_CANCELLED;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "adb3Ioctl: errno=%d (%s)\n", errno, strerror(errno));
        return ADB3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADB3_STATUS)err;
    }
  }
}

