/*
** File: admxrc3_platform.c  
** Project: ADMXRC3 API module library
** Purpose: Implements Linux-specific functions for ADMXRC3 API module library
**
** (C) Copyright Alpha Data 2010-2012
*/

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>

#include "admxrc3_platform.h"
#include <ioctl_dflinux.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const char g_deviceName[] = "admxrc3i";

ADMXRC3_STATUS
admxrc3OpenDevice(
  unsigned int index,
  bool_t bPassive,
  ADMXRC3_HANDLE* phAdmxrc3)
{
  char filename[32];
  int hDevice;

  snprintf(filename, ARRAY_LENGTH(filename), "/dev/%s%u", g_deviceName, index);
  hDevice = open(filename, bPassive ? O_RDONLY : O_RDWR);
  if (hDevice < 0) {
    switch (errno) {
    case EACCES:
      return ADMXRC3_ACCESS_DENIED;

    case ENXIO:
    case ENOENT:
      return ADMXRC3_DEVICE_NOT_FOUND;

    default:
      fprintf(stderr, "admxrc3OpenDevice: errno=%d (%s)\n", errno, strerror(errno));
      return ADMXRC3_UNEXPECTED_ERROR;
    }
  }
  *phAdmxrc3 = hDevice;
  return ADMXRC3_SUCCESS;
}

ADMXRC3_STATUS
admxrc3CloseDevice(
  ADMXRC3_HANDLE hAdmxrc3)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;

  /* Close the device handle. This prevents client code from using the handle to do further operations (such as mmap) */
  if (close(hAdmxrc3) < 0) {
    switch (errno) {
    case EBADF:
      status = ADMXRC3_INVALID_HANDLE;
      break;

    default:
      fprintf(stderr, "admxrc3CloseDevice: errno=%d (%s)\n", errno, strerror(errno));
      status = ADMXRC3_UNEXPECTED_ERROR;
      break;
    }
  }

  if (ADMXRC3_SUCCESS != status) {
    return status;
  }

  admxrc3UnmapAllRegions(hAdmxrc3);

  return ADMXRC3_SUCCESS;
}

#define MAX_IOCTL_SIZE (64)
#define IOCTL_SIZE_IN_SHIFT  (0)
#define IOCTL_SIZE_OUT_SHIFT (6)

ADMXRC3_STATUS
admxrc3Ioctl(
  ADMXRC3_HANDLE hAdmxrc3,
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  if (sizeIn >= MAX_IOCTL_SIZE || sizeOut >= MAX_IOCTL_SIZE) {
    fprintf(stderr, "admxrc3Ioctl: sizeIn (%u) or sizeOut (%u) greater than 64\n", sizeIn, sizeOut);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)code, ((int)sizeIn << IOCTL_SIZE_IN_SHIFT) | ((int)sizeOut << IOCTL_SIZE_OUT_SHIFT));
  err = ioctl(hAdmxrc3, ioctlCode, pIoctl);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        fprintf(stderr, "admxrc3Ioctl: EINPROGRESS returned for blocking operation\n");
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3Ioctl: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}

ADMXRC3_STATUS
admxrc3IoctlNonBlock(
  ADMXRC3_HANDLE hAdmxrc3,
  ADMXRC3_TICKET* pTicket, /* ignored */
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  if (sizeIn >= MAX_IOCTL_SIZE || sizeOut >= MAX_IOCTL_SIZE) {
    fprintf(stderr, "admxrc3IoctlNonBlock: sizeIn (%u) or sizeOut (%u) greater than 64\n", sizeIn, sizeOut);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)code | 0x40U, ((int)sizeIn << IOCTL_SIZE_IN_SHIFT) | ((int)sizeOut << IOCTL_SIZE_OUT_SHIFT));
  err = ioctl(hAdmxrc3, ioctlCode, pIoctl);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3IoctlNonBlock: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}

ADMXRC3_STATUS
admxrc3FinishNonBlock(
  ADMXRC3_HANDLE hAdmxrc3,
  ADMXRC3_TICKET* pTicket,
  bool_t bWait,
  void* pIoctl,
  unsigned int sizeOut)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, bWait ? (int)DFLINUX_IOCTLCODE_FINISHWAIT : (int)DFLINUX_IOCTLCODE_FINISH, ((int)sizeOut << IOCTL_SIZE_OUT_SHIFT));
  err = ioctl(hAdmxrc3, ioctlCode, pIoctl);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case EPIPE:
        return ADMXRC3_NONBLOCK_IDLE;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3FinishNonBlock: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}

ADMXRC3_STATUS
admxrc3IoctlCancel(
  ADMXRC3_HANDLE hAdmxrc3)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)DFLINUX_IOCTLCODE_CANCEL, 0);
  err = ioctl(hAdmxrc3, ioctlCode, 0);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        fprintf(stderr, "admxrc3IoctlCancel: EINPROGRESS returned for blocking operation\n");
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3IoctlCancel: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}
