/*
** File: clock.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for setting and getting clock
**          generator frequencies.
**
** (C) Copyright Alpha Data 2009-2010
*/

#include <df.h>
#include "device.h"
#include "clock.h"
#include <admxrc3/types.h>

static DfIoStatus
mapAcquireStatus(
  CoreAcquireStatus acqStatus)
{
  switch (acqStatus) {
  case CoreAcquireSuccess:
    return DfIoStatusSuccess;

  case CoreAcquireMustWait:
    return DfIoStatusError(ADMXRC3_MUST_WAIT);

  case CoreAcquireCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreAcquireRemoved:
  case CoreAcquireFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreAcquireInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreAcquireInvalid:
  case CoreAcquireUnexpectedError:
  case CoreAcquireWaiting:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

DfIoStatus
ioctlGetClockFrequency(
	Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize,
  unsigned int outSize)
{
  IOCTLS_ADMXRC3_GETCLOCKFREQUENCY* pIoctl = (IOCTLS_ADMXRC3_GETCLOCKFREQUENCY*)pBuffer;
  DfIoStatus status;
  CoreTicket ticket;
  CoreAcquireStatus acqStatus;
  CoreClockStatus clockStatus;
  unsigned int clockIndex;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize || sizeof(pIoctl->out) != outSize ) {
    return DfIoStatusInvalid;
  }

  clockIndex = pIoctl->in.clockIndex;
  if (clockIndex >= pDevCtx->coreInterface.pGetNumClockGen(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  ticket.count = 1;
  ticket.resources[0].code = CORE_RESOURCE_CLOCKGEN;
  ticket.resources[0].index = (uint8_t)clockIndex;
  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &ticket);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }
  acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, &ticket, 0);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }

  clockStatus = pDevCtx->coreInterface.pClockGetFreq(pDevCtx->pCoreContext, &ticket, clockIndex, &pIoctl->out.frequency);
  switch (clockStatus) {
  case CoreClockSuccess:
    status = DfIoStatusSuccess;
    break;
  case CoreClockInvalidIndex:
    status = DfIoStatusError(ADMXRC3_INVALID_INDEX);
    break;
  case CoreClockHardwareError:
    status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
    break;
  default:
    status = DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
    break;
  }

  pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &ticket);
  return status;
}

static void
clockProgCallback(
  CoreTicket* pCoreTicket,
  void* pCallbackContext /* unused */,  
  CoreClockStatus status)
{
  ClockGeneratorTicket* pTicket = (ClockGeneratorTicket*)pCoreTicket;

  pTicket->status = status;
  dfEventSignal(&pTicket->event);
}

DfIoStatus
ioctlSetClockFrequency(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize,
  unsigned int outSize)
{
#if defined(ADMXRC3_CHECK_FLAGS) && ADMXRC3_CHECK_FLAGS
  const uint32_t legalFlagsMask = ADMXRC3_SETCLOCK_TESTONLY | ADMXRC3_SETCLOCK_MAXIMUM | ADMXRC3_SETCLOCK_MINIMUM;
#endif
  IOCTLS_ADMXRC3_SETCLOCKFREQUENCY* pIoctl = (IOCTLS_ADMXRC3_SETCLOCKFREQUENCY*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  ClockGeneratorTicket* pTicket = NULL;
  CoreAcquireStatus acqStatus;
  CoreClockStatus clockStatus;
  unsigned int clockIndex;
  uint64_t frequency;
  uint32_t flags;
  CoreClockWord clockWord;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize || sizeof(pIoctl->out) != outSize ) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  clockIndex = pIoctl->in.clockIndex;
  frequency  = pIoctl->in.frequency;
  flags = pIoctl->in.flags;
  if (clockIndex >= pDevCtx->coreInterface.pGetNumClockGen(pDevCtx->pCoreContext)) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }
#if defined(ADMXRC3_CHECK_FLAGS) && ADMXRC3_CHECK_FLAGS
  if (flags & ~legalFlagsMask) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  if (flags & ADMXRC3_SETCLOCK_MINIMUM) {
    flags |= CoreClockMinimumFrequency;
  }
  if (flags & ADMXRC3_SETCLOCK_MAXIMUM) {
    flags |= CoreClockMaximumFrequency;
  }
  clockStatus = pDevCtx->coreInterface.pClockTestFreq(pDevCtx->pCoreContext, clockIndex, flags, frequency, frequency, frequency, &clockWord);
  switch (clockStatus) {
  case CoreClockSuccess:
    break;

  case CoreClockInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreClockOutOfRange:
    return DfIoStatusError(ADMXRC3_INVALID_FREQUENCY);

  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }

  pTicket = dfPoolAlloc(&pDevCtx->clockGenerator.ticketPool, ClockGeneratorTicket);
  if (NULL == pTicket) {
    return DfIoStatusError(ADMXRC3_NO_MEMORY);
  }

  if (!(flags & ADMXRC3_SETCLOCK_TESTONLY)) {
    pTicket->ticket.count = 1;
    pTicket->ticket.resources[0].code = CORE_RESOURCE_CLOCKGEN;
    pTicket->ticket.resources[0].index = (uint8_t)clockIndex;
    acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &pTicket->ticket);
    if (CoreAcquireSuccess != acqStatus) {
      status = mapAcquireStatus(acqStatus);
      goto done;
    }
    acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, &pTicket->ticket, 0);
    if (CoreAcquireSuccess != acqStatus) {
      status = mapAcquireStatus(acqStatus);
      goto done;
    }

    dfEventInit(&pTicket->event);
    clockStatus = pDevCtx->coreInterface.pClockProgram(pDevCtx->pCoreContext, &pTicket->ticket, clockIndex, &clockWord, clockProgCallback, NULL);
    if (CoreClockSuccess == clockStatus) {
      dfEventWait(&pTicket->event);
      clockStatus = pTicket->status;
    }

    switch (clockStatus) {
    case CoreClockSuccess:
      break;

    case CoreClockInvalidIndex:
      status = DfIoStatusError(ADMXRC3_INVALID_INDEX);
      break;

    case CoreClockHardwareError:
      status = DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
      break;

    case CoreClockOutOfRange:
      status = DfIoStatusError(ADMXRC3_INVALID_FREQUENCY);
      break;

    default:
      status = DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
      break;
    }

    pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &pTicket->ticket);
    dfEventUninit(&pTicket->event);
  }

done:
  if (NULL != pTicket) {
    dfPoolFree(&pDevCtx->clockGenerator.ticketPool, pTicket);
  }

  if (status == DfIoStatusSuccess) {
    pIoctl->out.actualFrequency = clockWord.frequency;
  }

  return status;
}
